from __future__ import annotations

from dataclasses import dataclass

from django.core.cache import cache
from django.http import HttpRequest

from .conf import conf

if conf.app_is_installed:
    from .models import Cart


@dataclass
class SessionStorage:
    """
    Save the cart data to the user session
    """

    request: HttpRequest
    session_key: str = conf.CART_SESSION_KEY

    def load(self) -> list[dict]:
        return self.request.session.get(self.session_key, [])

    def save(self, items: list[dict]) -> None:
        self.request.session[self.session_key] = items
        self.request.session.modified = True

    def clear(self) -> None:
        del self.request.session[self.session_key]
        self.request.session.modified = True


@dataclass
class DBStorage:
    """
    Save the cart data to the database, use the session for unauthenticated users
    """

    request: HttpRequest

    def load(self) -> list[dict]:
        data = SessionStorage(self.request).load()
        if not self.request.user.is_authenticated:
            return data
        cart, _ = Cart.objects.get_or_create(
            customer=self.request.user, defaults={"items": data}
        )
        return cart.items

    def save(self, items: list[dict]) -> None:
        if not self.request.user.is_authenticated:
            SessionStorage(self.request).save(items)
        else:
            Cart.objects.update_or_create(
                customer=self.request.user,
                defaults={"items": items},
            )

    def clear(self) -> None:
        if not self.request.user.is_authenticated:
            SessionStorage(self.request).clear()
        else:
            Cart.objects.filter(customer=self.request.user).delete()


@dataclass
class CacheStorage:
    """Use django cache backend to store cart details"""

    request: HttpRequest
    timeout: int = conf.CART_CACHE_TIMEOUT
    _cache_key: str = conf.CART_SESSION_KEY

    @property
    def _cart_id(self) -> str:
        id_ = (
            str(self.request.user.pk)
            if self.request.user.is_authenticated
            else str(self.request.session.session_key)
        )
        return f"{self._cache_key}-{id_}"

    def load(self) -> list[dict]:
        return cache.get(self._cart_id, [])

    def save(self, items: list[dict]) -> None:
        cache.set(self._cart_id, items, timeout=self.timeout)

    def clear(self) -> None:
        cache.delete(self._cart_id)



from __future__ import annotations

import itertools
from decimal import Decimal
from typing import Iterator, Type, TypeVar, Union, cast

from attrs import Factory, asdict, define, field
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.http import HttpRequest

from .conf import conf
from .protocols import Storage
from .utils import get_module

__all__ = ("Cart", "CartItem", "get_cart_class")

ProductModel = TypeVar("ProductModel", bound=models.Model)
Variant = Union[str, int, dict, set]


@define(kw_only=True)
class CartItem:
    quantity: int = field(eq=False, converter=int)
    variant: Variant | None = field(default=None)
    product_pk: str = field(converter=str)
    product_model_path: str
    metadata: dict = field(factory=dict, eq=False)

    @property
    def product(self) -> ProductModel:
        model = cast(Type[ProductModel], get_module(self.product_model_path))
        return model.objects.get(pk=self.product_pk)

    @property
    def subtotal(self) -> Decimal:
        return self.price * self.quantity

    @property
    def price(self) -> Decimal:
        return getattr(self.product, conf.CART_PRODUCT_GET_PRICE_METHOD)(self)

    @classmethod
    def from_product(
        cls,
        product: ProductModel,
        /,
        *,
        quantity: int,
        variant: Variant | None = None,
        metadata: dict | None = None,
    ) -> CartItem:
        metadata = metadata or {}
        return cls(
            quantity=quantity,
            variant=variant,
            product_pk=product.pk,
            product_model_path=f"{product.__class__.__module__}.{product.__class__.__name__}",
            metadata=metadata,
        )


@define(kw_only=True)
class Cart:
    request: HttpRequest
    storage: Storage
    _items: list[CartItem] = Factory(list)

    def __len__(self) -> int:
        return self.unique_count

    def __iter__(self) -> Iterator[CartItem]:
        yield from self._items

    def __contains__(self, item: CartItem) -> bool:
        return item in self._items

    @property
    def total(self) -> Decimal:
        return Decimal(sum(item.subtotal for item in self._items))

    @property
    def is_empty(self) -> bool:
        return self.unique_count == 0

    @property
    def count(self) -> int:
        """
        The number of items in the cart, that's the sum of quantities.
        """
        return sum(item.quantity for item in self._items)

    @property
    def unique_count(self) -> int:
        """
        The number of unique items in the cart, regardless of the quantity.
        """
        return len(self._items)

    @property
    def products(self) -> list[ProductModel]:
        """
        The list of associated products.
        """
        return [item.product for item in self._items]

    def find(self, **criteria) -> list[CartItem]:
        """
        Returns a list of cart items matching the given criteria.
        """

        def get_item_dict(item: CartItem) -> dict:
            return {key: getattr(item, key) for key in criteria}

        return [item for item in self._items if get_item_dict(item) == criteria]

    def find_one(self, **criteria) -> CartItem | None:
        """
        Returns the first cart item that matches the given criteria, if no match is found return None.
        """
        try:
            return self.find(**criteria)[0]
        except IndexError:
            return None

    def add(
        self,
        product: ProductModel,
        *,
        quantity: int = 1,
        variant: Variant | None = None,
        override_quantity: bool = False,
        metadata: dict | None = None,
    ) -> CartItem:
        """
        Add a new item to the cart

        :param product: An instance of a database product
        :param quantity: The quantity that will be added to the dj_shop_cart
        :param variant:  Variant details of the product
        :param override_quantity: If true will overwrite the quantity of the item if it already exists
        :param metadata: Optional metadata that is attached to the item, this dictionary can contain
            anything that you would want to attach to the created item in cart, the only requirements about
            it is that it needs to be json serializable
        :return: An instance of the item added
        """
        quantity = int(quantity)
        assert quantity >= 1, f"Item quantity must be greater than 1: {quantity}"
        item = self.find_one(product=product, variant=variant)
        if not item:
            item = CartItem.from_product(
                product, quantity=0, variant=variant, metadata=metadata
            )
            self._items.append(item)
        self.before_add(item=item, quantity=quantity)
        if override_quantity:
            item.quantity = quantity
        else:
            item.quantity += quantity
        self.after_add(item=item)
        self.save()
        return item

    def remove(
        self,
        product: ProductModel,
        *,
        quantity: int | None = None,
        variant: Variant | None = None,
    ) -> CartItem | None:
        """
        Remove an item from the cart entirely or partially based on the quantity

        :param product: An instance of a database product
        :param quantity: The quantity of the product to remove from the cart
        :param variant: Variant details of the product
        :return: The removed item with an updated quantity or None
        """
        item = self.find_one(product=product, variant=variant)
        self.before_remove(item=item, quantity=quantity)
        if not item:
            return None
        if quantity:
            item.quantity -= int(quantity)
        else:
            item.quantity = 0
        if item.quantity <= 0:
            self._items.pop(self._items.index(item))
        self.after_remove(item=item)
        self.save()
        return item

    def save(self) -> None:
        data = [asdict(item) for item in self._items]
        self.storage.save(data)

    def empty(self) -> None:
        self._items = []
        self.storage.clear()

    def variants_group_by_product(self) -> dict[str, list[CartItem]]:
        """
        Return a dictionary with the products ids as keys and a list of variant as values.
        """
        return {
            key: list(items)
            for key, items in itertools.groupby(
                self._items, lambda item: item.product_pk
            )
        }

    def before_add(self, item: CartItem, quantity: int) -> None:
        pass

    def after_add(self, item: CartItem) -> None:
        pass

    def before_remove(
        self, item: CartItem | None = None, quantity: int | None = None
    ) -> None:
        pass

    def after_remove(self, item: CartItem | None = None) -> None:
        pass

    @classmethod
    def new(cls, request: HttpRequest) -> Cart:
        """Appropriately create a new cart instance. This builder load existing cart if needed."""
        storage = get_module(conf.CART_STORAGE_BACKEND)(request)
        instance = cls(request=request, storage=storage)
        instance._items = [CartItem(**item) for item in storage.load()]
        return instance


def get_cart_class() -> type[Cart]:
    """
    Returns the correct cart manager class
    """
    klass = get_module(conf.CART_CLASS)
    if not issubclass(klass, Cart):
        raise ImproperlyConfigured(
            "The `CART_CLASS` settings must be a subclass of the `Cart` class."
        )
    return klass



from __future__ import annotations

from django.conf import settings


class Settings:
    """
    Shadow Django's settings with a little logic
    """

    @property
    def CART_SESSION_KEY(self) -> str:
        return getattr(settings, "CART_SESSION_KEY", "CART-ID")

    @property
    def CART_CACHE_TIMEOUT(self) -> int:
        # default to 5 days
        return getattr(settings, "CART_CACHE_TIMEOUT", 60 * 60 * 24 * 5)

    @property
    def CART_CLASS(self) -> str:
        return getattr(settings, "CART_CLASS", "dj_shop_cart.cart.Cart")

    @property
    def CART_PRODUCT_GET_PRICE_METHOD(self) -> str:
        return getattr(settings, "CART_PRODUCT_GET_PRICE_METHOD", "get_price")

    @property
    def CART_STORAGE_BACKEND(self) -> str:
        backend = getattr(settings, "CART_STORAGE_BACKEND", None)
        if backend == "dj_shop_cart.storages.DBStorage":
            assert (
                "dj_shop_cart" in settings.INSTALLED_APPS
            ), "You need to add dj_shop_cart to INSTALLED_APPS to use the DBStorage"
        return backend or "dj_shop_cart.storages.SessionStorage"

    @property
    def app_is_installed(self) -> bool:
        return "dj_shop_cart" in settings.INSTALLED_APPS


conf = Settings()



from __future__ import annotations

from django.conf import settings
from django.db import models
from django.utils import timezone


class Cart(models.Model):
    items = models.JSONField()
    customer = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    created = models.DateTimeField(default=timezone.now)
    modified = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return f"Cart - {self.customer}"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dj_shop_cart', 'dj_shop_cart.migrations', 'migrations']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.0', 'attrs>=21.4.0,<22.0.0']

setup_kwargs = {
    'name': 'dj-shop-cart',
    'version': '4.1.0',
    'description': 'Simple django cart manager for your django projects.',
    'long_description': '# dj-shop-cart\n\nA simple and flexible cart manager for your django projects.\n\n[![pypi](https://badge.fury.io/py/dj-shop-cart.svg)](https://pypi.org/project/dj-shop-cart/)\n[![python](https://img.shields.io/pypi/pyversions/dj-shop-cart)](https://github.com/Tobi-De/dj-shop-cart)\n[![django](https://img.shields.io/pypi/djversions/dj-shop-cart)](https://github.com/Tobi-De/dj-shop-cart)\n[![MIT License](https://img.shields.io/apm/l/atomic-design-ui.svg?)](https://github.com/Tobi-De/dj-shop-cart/blob/master/LICENSE)\n[![black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n✨📚✨ [Read the full documentation](https://tobi-de.github.io/dj-shop-cart/)\n\n## Features\n\n- Add, remove, decrement and clear items from cart\n- Authenticated users cart can be saved to database\n- Write custom methods to easily hook into the items add / remove flow\n- Custom **get_price** method to ensure that the cart always have an up-to-date products price\n- Each item in the cart hold a reference to the associated product\n- Metadata data can be attached to cart items\n- Supports specification of product variation details\n- Available context processor for easy access to the user cart in all your django templates\n- Swappable backend storage, with session and database provided by default\n\n\n## Installation\n\nInstall **dj-shop-cart** with pip or poetry.\n\n```bash\n  pip install dj-shop-cart\n```\n\n## Quickstart\n\n```python3\n\n# settings.py\n\nTEMPLATES = [\n    {\n        "OPTIONS": {\n            "context_processors": [\n                ...,\n                "dj_shop_cart.context_processors.cart", # If you want access to the cart instance in all templates\n            ],\n        },\n    }\n]\n\n# models.py\n\nfrom django.db import models\nfrom dj_shop_cart.cart import CartItem\nfrom decimal import Decimal\n\nclass Product(models.Model):\n    ...\n\n    def get_price(self, item:CartItem)->Decimal:\n        """The only requirements of the dj_shop_cart package apart from the fact that the products you add\n        to the cart must be instances of django based models. You can use a different name for this method\n        but be sure to update the corresponding setting (see Configuration). Even if you change the name the\n        function signature should match this one.\n        """\n\n\n# views.py\n\nfrom dj_shop_cart.cart import get_cart_class\nfrom django.http import HttpRequest\nfrom django.views.decorators.http import require_POST\n\n# This function has nothing to do with the package itself\nfrom .helpers import collect_params\n\nCart = get_cart_class()\n\n\n@require_POST\ndef add_product(request: HttpRequest):\n    product, quantity = collect_params(request)\n    cart = Cart.new(request)\n    cart.add(product, quantity=quantity)\n    ...\n\n\n@require_POST\ndef remove_product(request: HttpRequest):\n    product, quantity = collect_params(request)\n    cart = Cart.new(request)\n    cart.remove(product, quantity=quantity)\n    ...\n\n\n@require_POST\ndef empty_cart(request: HttpRequest):\n    Cart.new(request).empty()\n    ...\n\n```\n\n## Used By\n\nThis project is used by the following companies:\n\n- [Fêmy bien être](https://www.femybienetre.com/)\n\n## Development\n\nPoetry is required (not really, you can set up the environment however you want and install the requirements\nmanually) to set up a virtualenv, install it then run the following:\n\n```sh\npoetry install\npre-commit install --install-hooks\n```\n\nTests can then be run quickly in that environment:\n\n```sh\npytest\n```\n\n## Feedback\n\nIf you have any feedback, please reach out to me at degnonfrancis@gmail.com\n\n## Credits\n\nThanks to [Jetbrains](https://jb.gg/OpenSource) for providing an Open Source license for this project.\n\n<img height="200" src="https://resources.jetbrains.com/storage/products/company/brand/logos/jb_beam.png" alt="JetBrains Logo (Main) logo.">\n',
    'author': 'Tobi DEGNON',
    'author_email': 'tobidegnon@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Tobi-De/dj-shop-cart',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from importlib import import_module


def get_module(path: str) -> type:
    package, module = path.rsplit(".", 1)
    return getattr(import_module(package), module)
